//------------------------------------------------------------------------------
//
//  scene.cpp
//
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//  include
//------------------------------------------------------------------------------
#include "common.h"

static LPDIRECT3DPIXELSHADER9    pshader    = NULL;
static LPDIRECT3DPIXELSHADER9    pshader_f  = NULL;
static LPD3DXTEXTURESHADER       tshader    = NULL;
static LPDIRECT3DTEXTURE9        texture0   = NULL;
static LPDIRECT3DSURFACE9        surface0   = NULL;
static LPDIRECT3DSURFACE9        backbuffer = NULL;
static LPCSTR                    filename   = TEXT("main.fx");
static LPCSTR                    str_ps     = "ps_main";
static LPCSTR                    str_ps_f   = "ps_main_f";
static LPCSTR                    str_ts     = "ts_main";

int filter = 0;
static LPDIRECT3DTEXTURE9        noise     = NULL;
static LPDIRECT3DVOLUMETEXTURE9  noise3d   = NULL;


//shader constant
static D3DXVECTOR4 shader_const[256];

//------------------------------------------------------------------------------
//  variable
//------------------------------------------------------------------------------
struct d3dverts {
  FLOAT x, y, z, w;
  FLOAT u, v;
  enum { fvf = D3DFVF_XYZRHW | D3DFVF_TEX1 };
};

static const d3dverts quadverts[4] = {
 { -0.5+      0,  -0.5+      0, 0.0f, 1.0f, 0.0f, 0.0f},
 { -0.5+ScreenX,  -0.5+      0, 0.0f, 1.0f, 1.0f, 0.0f},
 { -0.5+      0,  -0.5+ScreenY, 0.0f, 1.0f, 0.0f, 1.0f},
 { -0.5+ScreenX,  -0.5+ScreenY, 0.0f, 1.0f, 1.0f, 1.0f},
};

//------------------------------------------------------------------------------
//
//  function 
//
//------------------------------------------------------------------------------


int mrand() {
  static int a = 1;
  static int b = 234567;
  static int c = 890123;
  a += b;b += c;c += a;
  return (a>>16);
}

float frand() {
  return float(mrand()) / 32767.9f;
}

static void WINAPI func_noise(D3DXVECTOR4* pOut, const D3DXVECTOR2* pTexCoord, const D3DXVECTOR2* pTexelSize, void* data) {
  pOut->x = frand();
  pOut->y = frand();
  pOut->z = frand();
  pOut->w = frand();
}


//------------------------------------------------------------------------------
//  InitScene
//------------------------------------------------------------------------------
void InitScene() {
  HRESULT hRet = E_FAIL;
  LPD3DXBUFFER tmp;
  LPD3DXBUFFER err;
  TermScene();
  DWORD option = D3DXSHADER_OPTIMIZATION_LEVEL3 | D3DXSHADER_PREFER_FLOW_CONTROL;
  
  //pass 1
  hRet = D3DXCompileShaderFromFile( filename,  0, 0, str_ps, "ps_3_0", option, &tmp, &err, 0 );
  if(hRet != D3D_OK) {
    printf("Error : %s\n", err->GetBufferPointer());
    return;
  }
  hRet = d3ddevice->CreatePixelShader( (DWORD*)tmp->GetBufferPointer(), &pshader );
  if(hRet != D3D_OK) {
    printf("Error : %s\n", err->GetBufferPointer());
    return;
  }
  RELEASE(tmp);
  
  //pass 2
  hRet = D3DXCompileShaderFromFile( filename,  0, 0, str_ps_f, "ps_3_0", option, &tmp, &err, 0 );
  if(hRet != D3D_OK) {
    printf("Error : %s\n", err->GetBufferPointer());
    return;
  }
  hRet = d3ddevice->CreatePixelShader( (DWORD*)tmp->GetBufferPointer(), &pshader_f );
  if(hRet != D3D_OK) {
    printf("Error : %s\n", err->GetBufferPointer());
    return;
  }
  RELEASE(tmp);
  
  //noise
  hRet = D3DXCompileShaderFromFile( filename,  0, 0, str_ts, "tx_1_0", option, &tmp, &err, 0 );
  if(hRet != D3D_OK) {
    printf("Error : %s\n", err->GetBufferPointer());
    return;
  }
  hRet = D3DXCreateTextureShader( (DWORD*)tmp->GetBufferPointer(), &tshader );
  if(hRet != D3D_OK) {
    printf("Error : %s\n", err->GetBufferPointer());
    return;
  }
  RELEASE(tmp);
  
  
  printf("Load Shader.\n");
  
  //Create Back Buffer
  d3ddevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &backbuffer);
  d3ddevice->CreateTexture(ScreenX, ScreenY, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &texture0, 0);
  texture0->GetSurfaceLevel(0, &surface0);
  
  //Create Noise Texture
  d3ddevice->CreateTexture(NoiseXY, NoiseXY, 1, 0, D3DFMT_A32B32G32R32F, D3DPOOL_MANAGED, &noise, 0);
  D3DXFillTexture(noise, func_noise, 0);
  d3ddevice->SetTexture(1, noise);
  d3ddevice->SetTexture(D3DVERTEXTEXTURESAMPLER0, noise);
  
  
  //Create 3d noise
  d3ddevice->CreateVolumeTexture(NoiseXYZ, NoiseXYZ, NoiseXYZ, 0, 0, D3DFMT_L16, D3DPOOL_MANAGED, &noise3d, 0);
  D3DXFillVolumeTextureTX(noise3d, tshader);
  d3ddevice->SetTexture(2, noise3d);
  //d3ddevice->SetTexture(D3DVERTEXTEXTURESAMPLER0, noise3d);
  
  d3ddevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_MIRROR);
  d3ddevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_MIRROR);
  
  /*
  d3ddevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  d3ddevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
  */
  d3ddevice->SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  d3ddevice->SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
  d3ddevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  d3ddevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
}

//------------------------------------------------------------------------------
//  TermScene
//------------------------------------------------------------------------------
void TermScene() {
  RELEASE(tshader);
  RELEASE(noise3d);
  RELEASE(noise);
  RELEASE(texture0);
  RELEASE(surface0);
  RELEASE(backbuffer);
  RELEASE(pshader);
  RELEASE(pshader_f);
}

//------------------------------------------------------------------------------
//  DoScene
//------------------------------------------------------------------------------
void DoScene() {
  if(GetAsyncKeyState(VK_F5)) {
    InitScene();
  }
  if(GetAsyncKeyState(VK_F6)) {
    filter = !filter;
  }
  if(pshader) {
    if(d3ddevice->BeginScene() == D3D_OK) {
      static unsigned long t = timeGetTime();
      unsigned long tt = timeGetTime();
      if(filter) {
        d3ddevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
        d3ddevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
        d3ddevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
        
        d3ddevice->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
        d3ddevice->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
        d3ddevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
        
        d3ddevice->SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
        d3ddevice->SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
        d3ddevice->SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
      } else {
        d3ddevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_NONE);
        d3ddevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_NONE);
        d3ddevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
        
        d3ddevice->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_NONE);
        d3ddevice->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_NONE);
        d3ddevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
        
        d3ddevice->SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_NONE);
        d3ddevice->SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_NONE);
        d3ddevice->SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
      }
      d3ddevice->SetPixelShaderConstantF(0, (FLOAT*)&shader_const, 64);
      d3ddevice->SetPixelShader( pshader );
      shader_const[0].x += (float)(tt - t) / 1000.0f;
      t = tt;
      d3ddevice->SetRenderTarget(0, surface0);
      d3ddevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, 0, 1.0f, 0);
      d3ddevice->SetFVF( d3dverts::fvf  );
      d3ddevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, quadverts, sizeof(d3dverts));

      d3ddevice->SetRenderTarget(0, backbuffer);
      d3ddevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, 0, 1.0f, 0);
      d3ddevice->SetFVF( d3dverts::fvf  );
      d3ddevice->SetTexture(0, texture0 );
      d3ddevice->SetPixelShader( pshader_f );
      d3ddevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, quadverts, sizeof(d3dverts));
    } else {
      printf("ς\n");
    }
    d3ddevice->EndScene();
    d3ddevice->Present( NULL, NULL, NULL, NULL );
  }
}

